﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Text;
using System.Xml;
using XmlSerializersTest.Serializers;

namespace XmlSerializersTest
{
    class SampleProcessor : IDisposable
    {
        private readonly IEnumerable<IXmlSerializer> _serializers;
        private readonly string _sourcePath;
        private readonly XmlWriter _output;

        public SampleProcessor(string sourcePath, IEnumerable<IXmlSerializer> serializers, XmlWriter output)
        {
            _serializers = serializers;
            _sourcePath = sourcePath;
            _output = output;
            WritePreamble();
        }

        public void Process(object obj)
        {
            Process(obj, obj.GetType());
        }

        public void Process(object obj, Type typeToDeserialize)
        {
            var properties = new SampleAttributes(obj);

            string sourceFile = Path.Combine(_sourcePath, GetSourceFileName(obj.GetType()));
            var source = ReadFileContent(sourceFile);
            _output.WriteStartElement("Sample");
            _output.WriteAttributeString("name", obj.GetType().Name);
            _output.WriteAttributeString("title", properties.Title);
            _output.WriteStartElement("Source");
            _output.WriteCData(source);
            _output.WriteEndElement();

            foreach (var serializer in _serializers)
            {
                if (properties.SerializersToSkip.ContainsKey(serializer.Name))
                {
                    SkipSerializer(serializer.Name, properties.SerializersToSkip[serializer.Name]);
                }
                else
                {
                    ProcessSerializer(obj, serializer, typeToDeserialize);
                }
            }

            _output.WriteEndElement();
        }

        public void Dispose()
        {
            _output.WriteEndElement();
        }

        private void WritePreamble()
        {
            _output.WriteStartElement("Samples");
        }

        private static string GetSourceFileName(Type type)
        {
            return type.Name + ".cs";
        }

        private static string ReadFileContent(string file)
        {
            using (var reader = new StreamReader(file))
            {
                return reader.ReadToEnd();
            }
        }

        private void ProcessSerializer(object obj, IXmlSerializer serializer, Type typeToDeserialize)
        {
            _output.WriteStartElement("Serializer");
            _output.WriteAttributeString("name", serializer.Name);
            string xml = RecordSerializationResults(obj, serializer);
            RecordDeserializationResults(xml, serializer, typeToDeserialize);
            _output.WriteEndElement();
        }

        private void SkipSerializer(string name, string reason)
        {
            _output.WriteStartElement("Serializer");
                _output.WriteAttributeString("name", name);

                _output.WriteStartElement("Serialized");
                    _output.WriteAttributeString("skipped", "true");
                    _output.WriteStartElement("Output");
                        _output.WriteCData(reason);
                    _output.WriteEndElement();
                _output.WriteEndElement();

                _output.WriteStartElement("Deserialized");
                    _output.WriteAttributeString("skipped", "true");
                _output.WriteEndElement();

            _output.WriteEndElement();
        }

        private string RecordSerializationResults(object obj, IXmlSerializer serializer)
        {
            string xml = "";
            _output.WriteStartElement("Serialized");
            
            try
            {
                xml = SerializeToString(obj, serializer);
                _output.WriteAttributeString("exception", "false");
                _output.WriteStartElement("Output");
                _output.WriteCData(xml);
                _output.WriteEndElement();
            }
            catch (Exception ex)
            {
                _output.WriteAttributeString("exception", "true");
                RecordException(ex);
            }

            _output.WriteEndElement();
            return xml;
        }

        private void WriteFragment(string xml)
        {
            if (string.IsNullOrEmpty(xml)) return;
            using (var textReader = new StringReader(xml))
            using (var xmlReader = XmlReader.Create(textReader))
            {
                _output.WriteNode(xmlReader, false);
            }
        }

        private void RecordException(Exception exception)
        {
            _output.WriteStartElement("Exception");
            _output.WriteAttributeString("type", exception.GetType().FullName);
            _output.WriteAttributeString("message", exception.Message);
            _output.WriteCData(exception.ToString());
            _output.WriteEndElement();
        }

        private static string SerializeToString(object obj, IXmlSerializer serializer)
        {
            var sb = new StringBuilder();
            var settings = new XmlWriterSettings
                { 
                    OmitXmlDeclaration = true, 
                    Encoding = Encoding.UTF8,
                    Indent = true
                };

            using (var xmlWriter = XmlWriter.Create(sb, settings))
            {
                serializer.Serialize(obj, xmlWriter);
                xmlWriter.Close();
                return sb.ToString();
            }
        }

        private void RecordDeserializationResults(string xml, IXmlSerializer serializer, Type typeToDeserialize)
        {
            _output.WriteStartElement("Deserialized");

            if (string.IsNullOrEmpty(xml))
            {
                _output.WriteAttributeString("skipped", "true");
            }
            else
            {

                try
                {
                    DeserializeFromString(xml, serializer, typeToDeserialize);
                    _output.WriteAttributeString("exception", "false");
                }
                catch (Exception ex)
                {
                    _output.WriteAttributeString("exception", "true");
                    RecordException(ex);
                }
            }

            _output.WriteEndElement();
        }

        private static void DeserializeFromString(string xml, IXmlSerializer serializer, Type typeToDeserialize)
        {
            using (var xmlReader = new XmlTextReader(new StringReader(xml)))
            {
                var unused = serializer.Deserialize(xmlReader, typeToDeserialize);
                if (unused == null)
                {
                    throw new ApplicationException("Deserializer returned null");
                }
            }
        }
    }
}
